/*
        MemView -=- Anzeige der DOS-internen Speicherbelegung

        Compiler: Microsoft C
        by Marcus Grber 1989-93

        Analyse der Speicherbelegung
*/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <bios.h>
#include <string.h>
#include <fcntl.h>

#include "mv.h"
#include "mv_tool.h"
#include "mv_blkls.h"


/*----------------------------------------------------------------------------*/
struct HandleDesc HandleInfo[255];      // Info ber EMS-Handles
struct HandleDesc XMSHdlInfo[255];      // Info ber XMS-Handles
unsigned NumHandles,NumXMSHdls;         // Anzahl der EMS/XMS-Handles

void far *XMSfunc;              /* Zeiger auf XMS entry point */
int HMAexist=0;                 /* HMA vorhanden? (default: nicht da) */
struct DIB far *DosInfo;        /* Zeiger auf Dos-Info-Block */

/*----------------------------------------------------------------------------*/
int getdrmajor(void)
/*
   Versuchen, die Hauptversionsnummer von DR-DOS herauszufinden
*/
{
        union REGS r;

        r.x.ax=0x4452;                  // Get DR-DOS version ID
        int86(0x21,&r,&r);
        if(r.x.cflag) return 0;         // Carry: kein DR-DOS
        switch(r.x.ax) {                // Bekannte Versionsnummern:
          case 0x1063: return 3;
          case 0x1065: return 5;
          case 0x1067: return 6;
        }
        return 0;
}

/*----------------------------------------------------------------------------*/
int getumblink(void)
{
        union REGS r;

        r.x.ax=0x5802;                  // Get umb link state
        int86(0x21,&r,&r);
        return r.x.ax;                  // UMB-Modus zurckgeben
}

void setumblink(unsigned link)
{
        union REGS r;

        r.x.ax=0x5803;                  // Set umb link state
        r.x.bx=link;                    // Gewnschter UMB-Modus
        int86(0x21,&r,&r);
}

/*----------------------------------------------------------------------------*/
void LoadPointers(void)
{
        union REGS r;
        struct SREGS segregs;
        int handle;

        FP_SEG(MyPSP)=_psp;             /* Eigener PSP liegt bei _psp:0000h */
        FP_OFF(MyPSP)=0x0000;

        r.h.ah=0x52;                    /* DOS-Info-Block-Adresse bestimmen */
        intdosx(&r,&r,&segregs);
        FP_OFF(DosInfo)=r.x.bx-4;       /* DIB liegt bei ES:BX-4 */
        FP_SEG(DosInfo)=segregs.es;

        if(!_dos_open("EMMXXXX0",O_RDONLY,&handle)) {
                                        /* Versuchen, EMM anzusprechen */
          _dos_close(handle);           /* Keine Handles verschwenden */

          r.h.ah=0x46;                  /* Get EMM version */
          int86(0x67,&r,&r);            /* EMS-Funktionsaufruf */
          if(!r.h.ah) EMSvers=r.h.al;   /* Erfolgreich: EMS-Version merken */
        }

        if(_dos_getvect(0x2F)) {        /* DOS 2.x: INT 2Fh zeigt ins Nichts */
          r.x.ax=0x4300;                /* XMS: Get install state */
          int86(0x2F,&r,&r);            /* XMS-Interface (Int) */
          if(r.h.al==0x80) {            /* Rckgabe "installiert"? */
            r.x.ax=0x4310;              /* XMS: Get entry point */
            int86x(0x2F,&r,&r,&segregs);/* XMS-Interface (Int) */
            FP_SEG(XMSfunc)=segregs.es; /* ES:BX zeigt auf XMS-Prozedur */
            FP_OFF(XMSfunc)=r.x.bx;
            r.h.ah=0x00;                /* Get XMS version */
            call86x(XMSfunc,&r,&r,&segregs);
                                        /* XMS-Interface (Prozedur) */
            XMSvers=r.x.ax;             /* Version merken */
            HMAexist=r.x.dx;            /* Flag "HMA existiert" */
          }

          if(drdosumb) {                // DR-DOS-kompatible UMBs?
            r.x.ax=0x12FF;              // DR-DOS-Infos zur Speicherverwaltung
            r.x.bx=0x0006;
            r.x.cx=r.x.dx=0;
            int86(0x2F,&r,&r);
            if(!r.x.cflag && !r.x.ax && r.x.dx)
                                        // UMB-Kette korrekt adressiert?
              moreMCBs[num_moreMCBs++]=r.x.dx;
                                        // Anfangsadresse der Kette merken
          }
        }
}

/*----------------------------------------------------------------------------*/
struct MCB far *LoadMCBstruc(unsigned seg,int noDOS)
{
        struct MemBlDesc blk;
        struct MCB far *ThisMCB;
        struct MCB far *ThisSCB;        /* DOS 4.0-SCBs aufgebaut wie MCBs */
        char ident;

        FP_SEG(ThisMCB)=seg;            /* Adresse des ersten MCBs */
        FP_OFF(ThisMCB)=0x0000;         /* Segment ist immer 0000h */
        do {
          ident=ThisMCB->ID;            /* Identifikation des MCBs holen */
          blk.seg=FP_SEG(ThisMCB)+1;    /* Segment des Blocks */
          blk.ofs=0x0000;               /* Block beginnt auf Segmentgrenze */
          blk.size=(long)(ThisMCB->LEN)<<4;
                                        /* Umfang des Blocks in Bytes */
          blk.psp=ThisMCB->PSP;         /* PSP des "Besitzers" */
          blk.sys=noDOS;                /* Zu welcher Kette gehrt der Block? */
          AddMemBlock(&FirstBlkDesc,&blk,0);
                                        /* Block in oberster Ebene einordnen */
          if(dos4scb && blk.psp==0x0008 && (_osmajor<5 ||
             (ThisMCB->OWNER[0]=='S' && ThisMCB->OWNER[1]=='D'))) {
                                        /* Ab 4.0: Systemblock evtl. mit SCBs */
            ThisSCB=ThisMCB;            /* 1.SCB beginnt 16 Bytes nach MCB */
            FP_SEG(ThisSCB)++;
            do {
              blk.type=ThisSCB->ID;     /* Identifikation des SCBs holen */
              blk.seg=FP_SEG(ThisSCB)+1;/* Segment des Blocks */
              blk.ofs=0x0000;           /* Block beginnt auf Segmentgrenze */
              blk.size=(long)(ThisSCB->LEN)<<4;
                                        /* Umfang des Blocks in Bytes */
              switch(blk.type) {        /* Sonderbehandlung fr einige Typen */
                case 'D':               /*** "D": Gertetreiber */
                case 'I':               /*** "I": IFS-Treiber (???) */
                  Get8Chars(blk.D_type.name,ThisSCB->OWNER);
                                        /* Treiber-Dateinamen aus SCB holen */
                  blk.D_type.mindrv=blk.D_type.maxdrv=(char)0xFE;
                                        /* Einheitennummern noch unbekannt */
              }
              AddMemBlock(&FirstBlkDesc,&blk,1);
                                        /* Block in zweiter Ebene einordnen */
              FP_SEG(ThisSCB)+=ThisSCB->LEN+1;
                                        /* Um Blockgre weiterschalten */
            } while(ThisSCB->ID!='M' && ThisSCB->ID!='Z');
                                        /* Bis auf MCB gestoen */
          }
          FP_SEG(ThisMCB)+=ThisMCB->LEN+1;
                                        /* Um Blockgre weiterschalten */
        } while(ident!='Z');            /* Bis MCB mit Endkennung gefunden */

        return ThisMCB;                 /* Zeiger auf letzten MCB zurckgeben */
}

/*----------------------------------------------------------------------------*/
void LoadMCBs(void)
{
        struct MemBlDesc blk,*p;
        struct MCB far *LastMCB,far *UMBchain;
        long ramfree;
        int i;
        unsigned link;
        unsigned bios_mem;

        bios_mem=_bios_memsize();       /* Was sagt's BIOS zum Hauptspeicher? */
        blk.seg=0x0070;                 /* Pseudo-Block ab 0000:0070 */
        blk.ofs=0x0000;
        blk.size=(long)(FP_SEG(DosInfo->PFAB)-0x0070)<<4;
                                        /* Umfang des Blocks in Bytes */
        blk.psp=0x0008;                 /* Besitzer ist das BIOS */
        blk.name[0]='\0';               /* Noch kein Name */
        blk.sys=0;                      /* Block in Systemkette */
        AddMemBlock(&FirstBlkDesc,&blk,0);
                                        /* Ersten (Pseudo-)Block eintragen */

        if(_osmajor>=5) {
          link=getumblink();            /* UMB-Linkstatus merken */
          setumblink(1);                /* UMBs zum Base Memory hinzufgen */
        }
        LastMCB=LoadMCBstruc(FP_SEG(DosInfo->PFAB),0);
                                        /* DOS-eigene MCB-Struktur lesen */
        if(_osmajor>=5)
          setumblink(link);             /* Alten UMB-Linkstatus wiederherst. */

        ramfree=1024L*bios_mem-_phys(LastMCB);
                                        /* KBytes in Bytes umrechnen, Platz
                                           letzter MCB-Block bis Ende */

        if(ramfree>0 && _osmajor>=5) {  /* Platz ber MCBs ab DOS 5.00:
                                           Wahrscheinlich Anfang UMB-MCBs */
          FP_SEG(UMBchain)=FP_SEG(LastMCB)+((unsigned)(ramfree>>4)-1);
          FP_OFF(UMBchain)=0x0000;      /* Letzter Paragraph unter RamTop */
          if(UMBchain->ID=='M') {       /* Korrekte Block-ID? */
            moreMCBs[num_moreMCBs++]=FP_SEG(UMBchain);
                                        /* Dann beginnt hier neue MCB-Kette */
            ramfree-=16;                /* 16 Bytes der Lcke sind geklrt */
          }
        }
        if(ramfree>0) {                 /* Noch RAM nach MCB-Struktur? */
          blk.seg=FP_SEG(LastMCB);      /* Segment des Blocks */
          blk.ofs=0x0000;               /* Block beginnt auf Segmentgrenze */
          blk.size=ramfree;             /* Umfang des Blocks in Bytes */
          blk.psp=0x0002;               /* Typ: unbekannter Block */
          blk.name[0]='r';              /* Kennzeichen: RAM ber MCBs */
          blk.sys=1;                    /* Block auerhalb aller MCBs */
          AddMemBlock(&FirstBlkDesc,&blk,0);
        }                               /* Block eintragen */

        if(!opt_keep) {                 /* Eigenen Block freigeben? */
          FreeBlock(FP_SEG(MyPSP));     /* Programmblock eigentlich frei */
          FreeBlock(MyPSP->ENV);        /* Environment eigentlich frei */
        }
        if(!(p=GetLocInfo(FirstBlkDesc,MyPSP,0)))
                                        /* Block mit eigenem Segment finden */
          return;                       /* Fehler: Abbrechen */
        MaxSize=p->size-(_phys(MyPSP)-_phys(p->beg));
                                        /* Abstand bis zum Ende des eigenen
                                           Segments */
        if((p=p->next) && !(p->psp))    /* Folgende freie Blcke dazuschlagen */
          MaxSize+=p->size+16;          /* Ein MCB weniger = 16 Bytes mehr */

        for(i=0;i<num_moreMCBs;i++)     /* Zustzliche MCB-Strukturen holen */
          if(!GetLocInfo(FirstBlkDesc,_ptr(moreMCBs[i],0x0010),0))
                                        /* Nur, wenn Block noch nicht bekannt */
            LoadMCBstruc(moreMCBs[i],i+1);
                                        /* Blockstrukturen einlesen */

        if(bios_mem<640) {              /* Weniger als 640 K Hauptspeicher? */
          blk.psp=0x0002;               /* Das ist uns einen Block wert... */
          blk.seg=bios_mem*64;          /* Segment des Blocks: nach BIOS-Mem */
          blk.ofs=0x0000;               /* Block beginnt auf Segmentgrenze */
          blk.size=(640-bios_mem)*1024; /* Umfang des Blocks in Bytes */
          blk.name[0]='n';              /* Kennzeichen: Kein RAM */
          blk.sys=1;                    /* Block auerhalb aller MCBs */
          AddMemBlock(&FirstBlkDesc,&blk,0);
        }
}

/*----------------------------------------------------------------------------*/
void GetPgName(struct MemBlDesc *blk)
/*
   Programmnamen holen und im Block (mu PSP-Block sein) ablegen. Es werden
   nur die ersten 8 Zeichen des Dateinamens gelesen.
*/
{
        unsigned char far *p,*p3;
        struct PSP far *pspptr;
        struct MemBlDesc *desc;
        struct MCB far *ThisMCB;
 
        pspptr=(struct PSP far *)blk->beg;
                                        /* Zeiger auf PSP */
        FP_SEG(ThisMCB)=blk->seg-1;     /* MCB liegt ein Segment vor Block */
        FP_OFF(ThisMCB)=0x0000;
        strcpy(blk->name,"?");          /* default: Name unbekannt */
        if(blk->psp==0x0008) {          /* BIOS-Datenblcke gesondert */
          if(_osmajor>=5 && _phys(blk->beg)!=0x700) {
                                        /* Ab DOS 5.0: ID fr MCB-BIOS-Daten */
            blk->name[0]=ThisMCB->OWNER[0];
            blk->name[1]=ThisMCB->OWNER[1];
            blk->name[2]='\0';          /* BIOS-Block-Typ aus MCB holen */
          }
          return;                       /* Name fertig */
        }
        if(pspptr->LDR==blk->seg && !dos4owner)
                                        /* Programm ohne Parent (COMMAND.COM) */
          if(p3=getenv("COMSPEC"))      /* Falls Kommandointerpreter-Name da */
            NameFromPath(blk->name,p3); /* Diesen als Programmnamen annehmen */
          else
            strcpy(blk->name,"COMMAND");/* Sonst Standard-Name "COMMAND" */
        else if(!dos4owner) {           /* Nicht DOS 4.0 kompatible MCBs */
          if(_osmajor>2) {              /* Name im Environment ab DOS 3.0 */
            FP_SEG(p)=pspptr->ENV;      /* Zeiger auf Environment-Anfang */
            FP_OFF(p)=0x0000;
            if((desc=GetBlkInfo(FirstBlkDesc,p,0))!=NULL && desc->psp==blk->seg) {
                                        /* Environment-Block noch da? */
              while(*p!='\0')           /* Bis Environment-Ende (Doppelnull) */
                while(*(p++)!='\0')     /* Jeweils stringweise weiter */
                  ;
              if(*(unsigned far *)(p+1)==0x0001)
                                        /* Identifikation fr Programmnamen */
                NameFromPath(blk->name,p+3);

                                        /* Namen aus Pfadnamen */
            }
          }
        }
        else                            /* Besitzername wie bei DOS 4.0: */
          Get8Chars(blk->name,ThisMCB->OWNER);
                                        /* Namen aus MCB holen */
}

/*----------------------------------------------------------------------------*/
void LoadPrograms(void)
{
        struct TechInfo info;
        struct MemBlDesc *p1,*p2;
        int i,files;
        struct PSP far *pspptr;

        p1=FirstBlkDesc;                /* Anfang der Speicherblockliste */
        while(p1!=NULL) {               /* Bis Listenende */
          p2=GetBlkInfo(FirstBlkDesc,pspptr=_ptr(p1->psp,0x0000),0);
                                        /* Besitzer-Block suchen */
          if((p2==NULL || p2->psp!=p1->psp) && p1->psp>0x0008)
                                        /* Nicht vorhanden? */
            p1->psp=0x0001;             /* Dann als fehlend kennzeichen */
          if((p1->psp==p1->seg && p1->psp!=0xFFFF) || p1->psp==0x0008
                                                   || p1->psp==0x0006)
            GetPgName(p1);              /* Benanntes Segment: Namen holen */
          if(p1->psp==p1->seg) {        /* Programm-Segment */
            files=0;                    /* Noch keine Dateien gefunden */
            for(i=pspptr->MAXFIL-1;i>4;i--)
                                        /* Alle Files (bis auf erste 5) durch */
              if(((char far *)(pspptr->PDOSIN))[i]!='\xFF')
                files++;                /* Eintrag fr Datei<>0: Datei offen */
            if(files) {                 /* Offene Dateien? */
              p2=GetLocInfo(FirstBlkDesc,pspptr->PDOSIN,0);
                                        /* Zeiger auf Block m. Handle-Tabelle */
              if(p2 && p2->psp==p1->psp) {
                                        /* Nur wenn Zeiger in gltigem Block */
                info.infotype=INFO_FILES;
                                        /* Info: Anzahl der Dateien */
                info.FILES.files=files; /* Anzahl der Dateien merken */
                AddTechInfo(p2,&info);  /* Technische Info zu Block zuordnen */
              }
            }
          }
          p1=p1->next;                  /* Einen Block weiter */
        }
}

/*----------------------------------------------------------------------------*/
void LoadDevices(void)
{
        struct MemBlDesc blk,*p;
        struct DEVHEAD far *devptr;
        struct DPB far *dpb;

        devptr=pDIBp(DosInfo,NULDEV);   /* Zeiger auf erste Einheit (NUL) */
        if(_fstrncmp(devptr->SPEC.NAM,"NUL     ",8))
          return;                       /* NUL-Device fehlt - keine Devices! */
        do {
          blk.type='D';                 /* Typ: Gertetreiber */
          blk.beg=devptr;               /* Anfagszeiger des Treibers */
          blk.size=-1;                  /* Lnge unbekannt */
          blk.D_type.name[0]='\0';      /* Kein Filename vorhanden */
          blk.D_type.mindrv=blk.D_type.maxdrv=(char)0xFE;
                                        /* Einheitennummern noch unbekannt */
          if(!(p=GetBlkInfo(FirstBlkDesc,devptr,1))
             || !(p->nopsp)
             || p->type!='D')           /* Treiber noch nicht registriert? */
            AddMemBlock(&FirstBlkDesc,&blk,1);
                                        /* Block in zweiter Ebene einordnen */
          devptr=devptr->LINK;          /* Zum nchsten Treiber */
        } while(FP_OFF(devptr)!=0xFFFF);/* Bis zum Ende der Kette */

        dpb=DosInfo->PDPB;              /* Zeiger auf ersten DPB */
        while(FP_OFF(dpb) && FP_OFF(dpb)!=0xFFFF) {
                                        /* Bis zum Ende der Kette */
          if(DPBp(dpb,DEV)) {           /* Gltiger Treiber? */
            p=GetBlkInfo(FirstBlkDesc,DPBp(dpb,DEV),1);
                                        /* Driver zu DPB suchen */
            if(p==NULL) {               /* Kein Treiber registriert: anlegen */
              blk.type='D';             /* Typ: Gertetreiber */
              blk.beg=DPBp(dpb,DEV);    /* Anfagszeiger des Treibers */
              blk.size=-1;              /* Lnge unbekannt */
              blk.D_type.name[0]='\0';  /* Kein Filename vorhanden */
              blk.D_type.mindrv=blk.D_type.maxdrv=(char)0xFE;
                                        /* Einheitennummern noch unbekannt */
              p=AddMemBlock(&FirstBlkDesc,&blk,1);
                                        /* Block in zweiter Ebene einordnen */
            }
            if(p!=NULL)                 /* ...sollte immer der Fall sein */
              if(p->D_type.mindrv==(char)0xFE)
                p->D_type.mindrv=p->D_type.maxdrv=dpb->DRIVE;
              else if(dpb->DRIVE<p->D_type.mindrv)
                p->D_type.mindrv=dpb->DRIVE;
              else if(dpb->DRIVE>p->D_type.maxdrv)
                p->D_type.maxdrv=dpb->DRIVE;
                                        /* Minimum-/Maximum-Laufwerk sammeln */
            dpb=DPBp(dpb,LINK);         /* Zum nchsten Parameterblock */
          }
        }
}

/*----------------------------------------------------------------------------*/
void GetEMSData(void)
{
        union REGS r;
        struct MemBlDesc blk,*p;
        struct {                        /* Liste der EMS mappable pages */
          unsigned seg;
          unsigned page;
        } EMSmaplist[64];
        struct {                        /* Handle-Directory (EMS 4.0, Fkt.21) */
          unsigned hdl;                 // Handle value
          char name[8];                 // Name des Handles
        } EMShandledir[255];
        unsigned i;

        if(EMSvers!=-1) {               /* Nur, falls EMS vorhanden */
          r.h.ah=0x41;                  /* Get Page Frame adress */
          int86(0x67,&r,&r);            /* EMS-Funktionsaufuruf */
          if(!r.h.ah) {                 /* Funktion erfolgreich */
            FP_SEG(blk.beg)=r.x.bx;     /* Segment des Pageframes */
            FP_OFF(blk.beg)=0x0000;
            blk.size=65536L;            /* EMS-Pageframe umfat immer 64K */
            blk.type='e';               /* Typ: Pageframe */
            AddMemBlock(&FirstBlkDesc,&blk,1);
          }                             /* Block eintragen */

          r.h.ah=0x42;                  /* Get Number of Pages */
          int86(0x67,&r,&r);            /* EMS-Funktionsaufruf */
          TotalEMS=r.x.dx*EMS_PAGE;     /* Gesamt-Speicherplatz unter EMS */
          FreeEMS=r.x.bx*EMS_PAGE;      /* Freier Speicherplatz unter EMS */

          DosEMS=0;                     /* Noch kein EMS unter DOS gefunden */
          NumHandles=0;                 /* Keine Handles in EMS<4.0 */
          if(EMSvers>=0x40) {           /* "Large Frame" erst ab LIM 4.0 */
            r.x.ax=0x5800;              /* Get adresses of mappable pages */
            r.x.di=(unsigned)EMSmaplist;/* Zeiger auf Rckgabepuffer */
            int86(0x67,&r,&r);          /* EMS-Funktionsaufruf */
            if(!r.h.ah)                 /* Funktion fehlerfrei? */
              for(i=0;i<r.x.cx;i++)     /* Alle mappable pages durchgehen */
                if(EMSmaplist[i].page>=4) {
                                        /* Pageframe interessiert nicht */
                  p=GetLocInfo(FirstBlkDesc,_ptr(EMSmaplist[i].seg,0x0010),0);
                                        /* Liegt page in bekanntem Block? */
                  if(p && !p->nopsp && p->psp!=0x0007)
                                        /* Ist es ein DOS-RAM-Block? */
                    DosEMS+=EMS_PAGE;   /* Kapazitt zusammenzhlen */
                }
            r.h.ah=0x4B;                /* Get handle count */
            int86(0x67,&r,&r);          /* EMS-Funktionsaufruf */
            if(!r.h.ah) {               /* Funktion fehlerfrei? */
              NumHandles=r.x.bx;        /* Anzahl der Handles holen */
              r.x.ax=0x5400;            /* Get handle directory */
              r.x.di=(unsigned)EMShandledir;
                                        /* Zeiger auf Rckgabepuffer */
              int86(0x67,&r,&r);        /* EMS-Funktionsaufruf */
              if(!r.h.ah)               /* Funktion fehlerfrei? */
                for(i=0;i<NumHandles;i++) {
                  HandleInfo[i].hdl=EMShandledir[i].hdl;
                  strncpy(HandleInfo[i].name,EMShandledir[i].name,8);
                  HandleInfo[i].name[8]='\0';
                                        /* Namen und Handlewert merken */
                  r.h.ah=0x4C;          /* Get handle pages */
                  r.x.dx=HandleInfo[i].hdl;
                  int86(0x67,&r,&r);    /* EMS-Funktionsaufruf */
                  HandleInfo[i].pages=(unsigned)( (!r.h.ah)? r.x.bx : 0);
                                        /* Anzahl der Seiten merken */
                  r.x.ax=0x5200;        /* Get handle attribute */
                  int86(0x67,&r,&r);    /* EMS-Funktionsaufruf */
                  HandleInfo[i].attr=(char)( (!r.h.ah)? r.h.al : -1);
                                        /* Attribut merken */
                }
              else
                NumHandles=0;           /* Vergessen wir's */
            }
          }
        }
}

/*----------------------------------------------------------------------------*/
int A20state(void)
/*
   Gibt 1 zurck, wenn Adreleitung A20 enabled ist (d.h. kein Wraparound bei
   Zugriff auf Segment FFFF:xxxx).
*/
{
        union REGS r;
        struct SREGS sr;

        r.h.ah=0x07;                    // Query A20 status
        call86x(XMSfunc,&r,&r,&sr);     // XMS-Interface
        return (r.h.bl==0)?r.x.ax:0;    // Fehler: A20 wohl nicht vorhanden
}

/*----------------------------------------------------------------------------*/
void GetXMSData(void)
{
        union REGS r;
        struct SREGS sr;
        struct MemBlDesc HMADesc,*p;
        struct TechInfo info;
        unsigned hdl;

        r.h.ah=0x88;                    /* Get Extended Memory size */
        int86(0x15,&r,&r);              /* AT-BIOS spezial */
        ExtTotal=r.x.ax*1024L;          /* Extended-Memory-Kapazitt */
        if(XMSvers!=-1) {               /* Nur, falls XMS vorhanden */
          if(HMAexist) {
            HMADesc.seg=0xFFFF;         /* FFFF = HMA-Kennzeichen */
            HMADesc.ofs=0x0000;
            HMADesc.size=65520;         /* Gre 64K - 16 Bytes */
            HMADesc.psp=0xFFFF;         /* default: Gehrt einem Programm */
            strcpy(HMADesc.name,"?");   /* Name des belegenden Programms: ??? */
            HMADesc.sys=1;              /* Wird zur oberen DOS-Kette gerechnet */
            r.h.ah=0x01;                /* Allocate HMA */
            r.x.dx=0xFFFF;              /* Applikation: Bekommt immer HMA */
            call86x(XMSfunc,&r,&r,&sr); /* XMS-Interface */
            if(r.x.ax==0x0001) {        /* HMA wurde belegt? */
              HMADesc.psp=0x0000;       /* HMA als frei kennzeichnen */
              r.h.ah=0x02;              /* Free HMA */
              call86x(XMSfunc,&r,&r,&sr);
                                        /* XMS-Interface */
            }
            else                        /* HMA ist schon belegt */
              if(_osmajor>=5) {         /* Unter DOS 5.0 prfen, ob belegt */
                r.x.ax=0x3306;          /* undok. Funktion */
                int86(0x21,&r,&r);      /* MS-DOS-Systemfunktion */
                if(r.x.dx & 0x1000)     /* Flag fr "MS-DOS in HMA"? */
                  HMADesc.psp=0x0008;   /* Gesetzt: HMA als System-Block */
              }
            p=AddMemBlock(&FirstBlkDesc,&HMADesc,0);
                                        // HMA wie PSP-Block in Kette */
            info.infotype=INFO_A20;     // HMA-Status als technisches Info dazu
            info.A20.state=A20state();  // A20-Status bestimmen
            AddTechInfo(p,&info);       // Technische Info dem HMA-Blk zuordnen
          }

          r.h.ah=0x08;                  /* Query free XMS memory */
          call86x(XMSfunc,&r,&r,&sr);   /* XMS-Interface */
          FreeXMS=r.x.dx*1024L;         /* Gesamter freier XMS-Speicher */

          NumXMSHdls=0;                 // Noch keine Handles gefunden
          if(opt_xmem) {                // Dauer zu lange, um's immer zu machen
            hdl=0;                      // Alle Handles ausprobieren
            do {
              r.h.ah=0x0E;              // Get EMB handle information
              r.x.dx=hdl;               // Handle ausprobieren
              call86x(XMSfunc,&r,&r,&sr); // XMS-Interface
              if(r.x.ax) {              // Handle existiert?
                XMSHdlInfo[NumXMSHdls].hdl=hdl;
                XMSHdlInfo[NumXMSHdls].pages=r.x.dx;
                XMSHdlInfo[NumXMSHdls].attr=(char)r.h.bh;
                                        // Handle-Daten speichern
                NumXMSHdls++;           // Handles zhlen
              }
            } while(++hdl && NumXMSHdls<255);
                                        // Bis alle Handles einmal durch
          }
        }
}

/*----------------------------------------------------------------------------*/
void LoadROMs(void)
{
        struct ROMHeader huge *ExtScan;
        struct MemBlDesc blk,*p,*p2;
        long blkend;
    
        FP_SEG(ExtScan)=0xC000;         /* Anfang fr Karten-BIOS-Suche */
        FP_OFF(ExtScan)=0x0000 ;        /* bei C000:0000 */
        for(;_phys(ExtScan)<0xF0000;ExtScan++)
                                        /* Bis Segment EFFFh scannen */
          if(ExtScan->ID==0xAA55 ||     /* Karten-ROM? (55AA/5555: Portfolio) */
             ((ExtScan->ID==0x55AA || ExtScan->ID==0x5555) && _osmajor==2)){
            blk.beg=ExtScan;            /* Segment des Blocks */
            blk.size=512L*ExtScan->size;/* Umfang des Blocks in Bytes */
            blk.type='x';               /* Typ: BIOS-ROM */
            AddMemBlock(&FirstBlkDesc,&blk,1);
            ExtScan+=ExtScan->size-1;   /* Zur Position nach dem ROM-Block */
          }                             /* Block eintragen */

        p2=NULL;                        /* Noch keinen Block gefunden */
        for(p=FirstBlkDesc;p && p->seg!=0xFFFF;p=p->next)
                                        /* Blcke abklappern */
          p2=p;                         /* Vorhergehenden Block merken */
        if(!p2 || (blkend=_phys(p2->beg)+p2->size)<=0xF0000)
                                        /* Kein Block reicht in Seg ab F000? */
          blk.beg=_ptr(0xF000,0);       /* Dann BIOS-ROM ab F000 ansetzen */
        else
          blk.beg=_ptr((unsigned)((blkend+0x1FFF)>>4)&0xFE00,0);
                                        /* Sonst ab F200,F400,F600,F800... */
        blk.size=0x100000-_phys(blk.beg);
                                        /* Lnge: Bis Obergrenze 1. Megabyte */
        blk.type='r';                   /* Typ: BIOS-ROM */
        AddMemBlock(&FirstBlkDesc,&blk,1);
                                        /* Block eintragen */
}

/*----------------------------------------------------------------------------*/
void LoadVideo(void)
{
        union REGS r;
        struct MemBlDesc blk;
        char mode;
	int vidA000,vidB000=0,vidB800=0;

        r.h.ah=0x0F;                    // Get video mode
        int86(0x10,&r,&r);              // Video-BIOS
        mode=(char)(r.h.al & 0x7F);     // Videomodus identifizieren
        if(mode!=0x07)                  // Nicht Monochrom-Modus?
          vidB800=1;                    // Dann ist wohl ein CGA-Puffer da
        else
          vidB000=1;                    // Sonst MDA/HGC-Puffer annehmen

        r.h.ah=0x12; r.h.bl=0x10;       // EGA-Info holen
        int86(0x10,&r,&r);              // Video-BIOS
        vidA000=(r.h.bl!=0x10);         // Wert in BL gendert - es ist EGA/VGA

        if(vidA000) {                   // EGA/VGA-Puffer
          blk.beg=_ptr(0xA000,0);       // 64 KByte ab A000h
          blk.size=65536;
          blk.type='v';                 // Videospeicher
          AddMemBlock(&FirstBlkDesc,&blk,1);
                                        // Block eintragen
        }
        if(vidB000) {                   // MDA/HGC-Puffer
          blk.beg=_ptr(0xB000,0);       // 32 KByte ab B000h
          blk.size=32768;
          blk.type='v';                 // Videospeicher
          AddMemBlock(&FirstBlkDesc,&blk,1);
                                        // Block eintragen
        }
        if(vidB800) {                   // CGA-Puffer
          blk.beg=_ptr(0xB800,0);       // 32 KByte ab B800h
          blk.size=32768;
          blk.type='v';                 // Videospeicher
          AddMemBlock(&FirstBlkDesc,&blk,1);
                                        // Block eintragen
        }
}

/*----------------------------------------------------------------------------*/
#define IRET 0xCF                       /* Interrupt-Rckkehr */

void LoadInterrupts(void)
{
        struct TechInfo info;
        struct MemBlDesc *p;
        unsigned intnr;                 /* Interrupt-Nummer */
        void far *intvec;               /* Interrupt-Vektor */
        int dir,count;

        for(intnr=0;intnr<(unsigned)(opt_userints?0xFF:0xBF);intnr++) {
                                        /* Interrupts durchgehen */
          intvec=(void far *)_dos_getvect(intnr);
                                        /* Interrupt-Vektor lesen */
          dir=1;                        /* Erste Referenz ist "direkt" */
          count=255;                    /* Nur maximal 255 Programme tracen */
          do {
            if(*(unsigned char far *)intvec==(unsigned char)IRET && !opt_irets)
              continue;                 /* berspringen, falls auf IRET */
            if((p=GetLocInfo(FirstBlkDesc,intvec,127))!=NULL) {
                                        /* Zieladresse mglichst exakt finden */
              info.infotype=INFO_INT|intnr;
                                        /* Interrupt-Nummer in Block */
              info.INT.direct=dir;      /* Vektor zeigt direkt auf Interrupt */
              AddTechInfo(p,&info);     /* Technische Info zu Block zuordnen */
            }
            dir=0;                      /* Nchster ist nicht mehr direkt */

/* Verschiedene "Tunneling"-Strategien: */

/* IBM interrupt sharing protocol / AMIS proposal by Ralf Brown: */
            if(FP_OFF(intvec)<=0xFFE0 &&
               ((struct IBMINTHEAD far *)intvec)->JMP==IBMINTHEAD_JMP &&
               ((struct IBMINTHEAD far *)intvec)->MAGIC==IBMINTHEAD_MAGIC) {
                                        /* Kein Segment-Wrap! */
              intvec=((struct IBMINTHEAD far *)intvec)->chain;
                                        /* Verkettung zum nchsten Interrupt */
            }
            else intvec=NULL;           /* Kein Nachfolger gefunden */

          } while(intvec && --count);
        }
}

/*----------------------------------------------------------------------------*/
void LoadTesSeRact(void)
{
        union REGS r,ro;
        struct SREGS sr;
        struct TechInfo info;
        struct MemBlDesc *p;
        struct TsrParms far *par;
        unsigned i;

        if(!_dos_getvect(0x2F))         /* INT 2Fh prfen */
          return;                       /* DOS 2.x: INT 2Fh zeigt auf NULL */
        r.x.ax=0x5453;                  /* TesSeRact-Identifikation */
        r.x.bx=0x01;                    /* Funktion 01h: Zeiger auf Daten */
        r.x.cx=0;                       /* Mgliche ID-Nummern abklappern */
        do {
          int86x(0x2F,&r,&ro,&sr);      /* Multiplex-Interrupt aufrufen */
          if(!ro.x.ax) {                /* TSR gefunden? */
            FP_SEG(par)=sr.es;          /* Zeiger auf TesSeRact-Parameter */
            FP_OFF(par)=ro.x.bx;
            info.infotype=INFO_TESN;    /* TesSeRact-Name */
            info.TESN.par=par;          /* Zeiger auf Parameter merken */
            if(p=GetLocInfo(FirstBlkDesc,par,127)) {
                                        /* Zieladresse mglichst exakt finden */
              AddTechInfo(p,&info);     /* Technische Info zu Block zuordnen */
              info.infotype=INFO_TESH;  /* TesSeRact-Hotkey */
              info.TESH.Shift=par->ShiftState;
              info.TESH.Hotkey=par->HotKey;
                                        /* Tastencode merken */
              AddTechInfo(p,&info);     /* Technische Info zu Block zuordnen */
              for(i=0;i<par->ExtraHotCount;i++) {
                                        /* Zustzliche Hotkeys ergnzen */
                info.TESH.Shift=
                  ((struct ExtraHot far *)(par->ExtraHotKeys))[i].ShiftState;
                info.TESH.Hotkey=
                  ((struct ExtraHot far *)(par->ExtraHotKeys))[i].Hotkey;
                AddTechInfo(p,&info);   /* Technische Info zu Block zuordnen */
              }
            }
          }
          r.x.cx++;                     /* Nchste Identifikationsnummer */
        } while(!ro.x.ax);              /* Bis zur ersten freien ID-Nummer */
}
