/*
        MemView -=- Anzeige der DOS-internen Speicherbelegung

        Compiler: Microsoft C
        by Marcus Grber 1989-93

        Anzeige der Speicherbelegung
*/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <conio.h>

#include "mv.h"
#include "mv_tool.h"
#include "mv_blkls.h"

unsigned _effadr(struct MemBlDesc *p)
/*
   Berechnet die effektive Adresse (=Sortierwert) fr den bergebene Block.
*/
{
        unsigned adr;

        if(p->nopsp)
          return (unsigned)(_phys(p->beg)>>4);
        adr=(p->psp)?(p->psp):0xFFFF;   // Freie Blcke ganz an Schlu
        if(adr==0x0006) return 0xFFFE;  // Davor DR-DOS XMS-UMBs
        if(adr==0x0007) return 0xFFFD;  // Davor DR-DOS excluded
}

int sortcmp(struct MemBlDesc *p1,struct MemBlDesc *p2)
/*
   Diese Funktion legt die Sortierreihenfolge fr die Blcke p1 und p2 fest.
   Der Rckgabewert entspricht der von strcmp() verwendeten Codierung.
*/
{
        unsigned adr1,adr2;

        adr1=_effadr(p1);               // Sortierwerte berechnen
        adr2=_effadr(p2);
        return adr1==adr2? 0:           // Rckgabewert wie strcmp()
                adr1<adr2?-1:
                          +1;
}

void SortByProgram(void)
/*
   Blcke der obersten Ebene (MCB-Blcke) nach Programmen (=PSPs) sortieren.
   Da die Anzahl der Blcke sich in Grenzen hlt, wird "nur" ein Bubblesort-
   Verfahren benutzt.
*/
{
        int flag;
        register struct MemBlDesc *p1,*p2,*temp;

        do {
          flag=0;                       /* default: Datei ist sortiert */
          p1=NULL;                      /* Noch kein Vorgnger */
          for(p2=FirstBlkDesc;p2!=NULL && p2->next!=NULL;p2=p2->next) {
                                        /* Ersten bis vorletzten Block durch */
           if(sortcmp(p2,p2->next)>0) { /* Sortieren nach Sortierwerten */
              if(p1!=NULL)              /* Verkettung im Vorgnger/Listenanf. */
                p1->next=p2->next;      /* auf Nachfolger setzen */
              else
                FirstBlkDesc=p2->next;
              temp=p2->next->next;      /* Zeiger auf bernchsten merken */
              p2->next->next=p2;        /* Nachfolger wird Vorgnger */
              p2->next=temp;
              flag=1;                   /* Mglicherweise noch unsortiert */
            }
            p1=p2;                      /* Letzten Block merken */
          }
        } while(flag);                  /* Bis sortiert */
}

/*-----------------------------------------------------------------------------*/
#define info_tab 39
#define vline_tab 22

#define NUMKEY 0x59             /* Anzahl der Tasten */
char *keys[NUMKEY]={            /* Scancodes -> Tasten-Namen */
  ""   ,"Esc","1"  ,"2"  ,"3"  ,"4"  ,"5"  ,"6"  ,"7"  ,"8"  ,   /* 00h-09h */
  "9"  ,"0"  ,""   ,""   ,"Bsp","Tab","Q"  ,"W",  "E"  ,"R"  ,   /* 0Ah-13h */
  "T"  ,"Y"  ,"U"  ,"I"  ,"O"  ,"P"  ,""   ,""   ,"Ret",""   ,   /* 14h-1Dh */
  "A"  ,"S"  ,"D"  ,"F"  ,"G"  ,"H"  ,"J"  ,"K"  ,"L"  ,""   ,   /* 1Eh-27h */
  ""   ,""   ,""   ,""   ,"Z"  ,"X"  ,"C"  ,"V"  ,"B"  ,"N"  ,   /* 28h-31h */
  "M"  ,""   ,""   ,"gr/",""   ,"gr*",""   ,"Spc",""   ,"F1" ,   /* 32h-3Bh */
  "F2" ,"F3" ,"F4" ,"F5" ,"F6" ,"F7" ,"F8" ,"F9" ,"F10",""   ,   /* 3Ch-45h */
  ""   ,"Hom","Up" ,"PgU","gr-","Lft","gr5","Rig","gr+","End",   /* 46h-4Fh */
  "Dwn","PgD","Ins","Del",""   ,""   ,""   ,"F11","F12"          /* 50h-59h */
};
char *shift[4]={"ShiftR","ShiftL","Ctrl","Alt"};        /* Shift-Tasten-Namen */

void ListTechInfo(struct TechInfo *info,int vline)
{
        char s[40],sp[40],sd[40];       /* Stringpuffer fr Infos */
        unsigned prev;                  /* Vorhergehende Info */
        unsigned len;                   /* Zeilenlnge */
        unsigned firstint;              /* Evtl. Interrupt-Reihen */
        unsigned type;                  /* Info-Typ */
        unsigned sub;                   /* Unterinfo-Typ */
        char buf[9];                    /* Puffer fr ID-Name */
        int i;
        int fl,m;

        prev=0x0000;                    /* Noch keine Info gelesen */
        len=0;                          /* Noch nichts ausgegeben */
        firstint=0xFFFF;                /* keine Interrupt-Reihe */
        for(;info!=NULL;info=info->next) {
                                        /* Alle Infos durchgehen */
          type=info->infotype & INFO_MASK;
          sub=info->infotype & INFO_SUB;/* Info aufspalten */
          switch(type) {
          /*------------------------------------------------------------------*/
            case INFO_INT:              /*** Interrupts */
              if(info->next!=NULL &&
                 info->next->infotype==info->infotype+1 &&
                 info->next->INT.direct==info->INT.direct) {
                                        /* Folgt noch ein Interrupt? */
                if(firstint==0xFFFF)    /* Neue Folge? */
                  firstint=sub;         /* Dann Anfangsglied merken */
                continue;               /* Nchste Info */
              }
              if(firstint==0xFFFF)      /* Nicht in Folge? */
                sprintf(s,info->INT.direct?"%02X":"(%02X)",sub);
              else {
                sprintf(s,info->INT.direct?"%02X-%02X":"(%02X-%02X)",
                        firstint,sub);  /* Entsprechenden String erzeugen */
                firstint=0xFFFF;        /* Folge erledigt */
              }
              sprintf(sp,"INT ");       /* Vorsatz bei Zeilenwechsel */
              sprintf(sd,",");          /* Vorsatz bei Forsetzung */
              break;
          /*------------------------------------------------------------------*/
            case INFO_TESS:             /*** TesSeRact */
              sprintf(sp,"Tess: ");     /* Vorsatz bei Zeilenwechsel */
              sprintf(sd," ");          /* Vorsatz bei Forsetzung */
              switch(sub) {             /* Unterinfos auflsen */
                case INFO_TESN & INFO_SUB:
                  _fstrncpy(buf,info->TESN.par->IdCode,8);
                                        /* ID-String ins lokale Segment holen */
                  for(i=8;i>0 && buf[i-1]==' ';i--)
                    ;                   /* Blanks am Ende entfernen */
                  buf[i]='\0';          /* Endmarke */
                  i=info->TESN.par->TsrStatus;
                                        /* TSR-Statusflags */
                  sprintf(s,"%c%s%c(%xh)",
                               (i & 0x400)? '^':
                               (i & 0x200)? '"':
                                            '(',
                               buf,     /* TesSeRact-Namen anzeigen */
                               (i & 0x400)? '^':
                               (i & 0x200)? '"':
                                            ')',
                               info->TESN.par->IdNum);
                                        /* TesSeRact-Nummer anzeigen */
                  break;                
                case INFO_TESH & INFO_SUB:
                  strcpy(s,"[");        /* Anfang des Strings */
                  if(info->TESH.Hotkey<NUMKEY &&
                     keys[info->TESH.Hotkey][0]!='\0') {
                                        /* Gltiger Tastencode? */
                    fl=0;               /* Noch keine Tasten */
                    for(i=0,m=1;i<4;i++,m<<=1)
                                        /* Shift-Tastenflags durch */
                      if(m & info->TESH.Shift) {
                                        /* Entsprechende Taste aktiv? */
                        if(fl)          /* Schon andere Tasten davor? */
                          strcat(s,"-");/* Dann Trennzeichen */
                        strcat(s,shift[i]);
                                        /* Tastensymbol ablegen */
                        fl=1;           /* Setzen: Schon andere Symbole */
                      }
                    if(info->TESH.Hotkey>0) {
                                        /* Tastencode definiert? */
                      if(fl)            /* Schon andere Tasten davor? */
                        strcat(s,"-");  /* Dann Trennzeichen */
                      strcat(s,keys[info->TESH.Hotkey]);
                                        /* Tastencode hinzufgen */
                    }
                  }
                  else
                    strcat(s,"?");      /* Unbekannter Tastencode */
                  strcat(s,"]");        /* Ende des Tastensymbols */
                  break;
              }
              break;
          /*------------------------------------------------------------------*/
            case INFO_FILES:            /*** Offene Dateien vorhanden */
              sprintf(sp,"Offene Dateien: ");
                                        /* Vorsatz bei Zeilenwechsel */
              sprintf(s,"%d",info->FILES.files);
                                        /* Anzahl der Dateien */
              break;
          /*------------------------------------------------------------------*/
            case INFO_A20:              /*** A20-Status */
              sprintf(sp,"A20-Status: ");
              sprintf(s,(info->A20.state)?"enabled":"disabled");
              break;
          /*------------------------------------------------------------------*/
          }
          if(len+strlen((prev!=type)?sp:sd)+strlen(s)+info_tab+
             ((prev!=type && prev)?4:0)>=opt_width) {
                                        /* Zeile voll? */
            _cprintf("\r\n%*s%c%*s",vline_tab,"",vline?'':' ',
                                    info_tab-vline_tab-1,"");
                                        /* Dann neue Zeile/tabulieren */
            len=0;                      /* Zeile noch leer */
          }
          else
            if(prev!=type && prev) {    /* Gruppenwechsel? */
              _cprintf("  ");         /* Trennzeichen */
              len+=4;                   /* Zeile verlngert */
            }
          _cprintf("%s",(prev!=type || len==0)?sp:sd);
                                        /* Zeilenvorsatz/Trenner ausgeben */
          _cprintf("%s",s);             /* Info ausgeben */
          len+=strlen((prev!=type || len==0)?sp:sd)+strlen(s);
          prev=type;                    /* Infotyp merken */
        }
}

/*----------------------------------------------------------------------------*/
int ListBlock(struct MemBlDesc **Bl,struct MemBlDesc **sub,int level)
{
        unsigned t_adr;
        long t_bytes;
        char *t_name,*t_desc,t_chain;
        struct TechInfo *info;
        int showadr;

        unsigned psp;
        struct PSP far * pspptr;
        struct DEVHEAD far * devptr;
        struct MemBlDesc *p2;
        void far *p;
        int count;
        char s[17],s2[9];

        t_adr=_normal((*Bl)->beg);      /* Anfangssegment des Blocks */
        t_chain=' ';                    /* Standardsymbol fr MCB-Kette */
        t_bytes=(*Bl)->size;            /* Umfang des Blocks */
        t_name=t_desc="";               /* keine Bezeichnung */
        (*sub)=(*Bl)->sub;              /* Zeiger Unterblcke/Infos */
        info=(*Bl)->info;
        showadr=1;                      /* default: Anfangssegment anzeigen */

        if((*Bl)->nopsp) {              /* Block je nach Typ anzeigen */
          switch((*Bl)->type) {         /* Blocktyp auswerten */
          /*------------------------------------------------------------------*/
            case 'D':                   /*** "D": Gertetreiber */
              devptr=(*Bl)->beg;        /* Zeiger auf Gertetreiber */
              if(devptr->ATTR & ATTR_CHARDEV) {
                                        /* Zeicheneinheit? */
                Get8Chars(s2,devptr->SPEC.NAM);
                                        /* Ja: Einheitennamen holen */
                sprintf(s,"Gert %s",s2); /* Einheitennamen als Beschreibung */
              }
              else                      /* Blockeinheit: */
                if((*Bl)->D_type.mindrv==(*Bl)->D_type.maxdrv)
                                        /* Nur ein Untertreiber? */
                  sprintf(s,"Laufwerk %c:",(*Bl)->D_type.mindrv+'A');
                else                    /* Mehrere Untertreiber */
                  sprintf(s,"Laufwerk %c:-%c:",
                             (*Bl)->D_type.mindrv+'A',(*Bl)->D_type.maxdrv+'A');
              t_name=(*Bl)->D_type.name;/* Unter DOS 4.0 steht hier Dateiname */
              t_desc=s;                 /* Beschreibung steht im Puffer */
              break;
          /*------------------------------------------------------------------*/
            case 'I':                   /*** "I": IFS-Treiber (nie gesehen!) */
              t_desc="IFS=";            /* Bezeichnung DOS 4.0-MEM-kompatibel */
              t_name=(*Bl)->D_type.name;/* Unter DOS 4.0 steht hier Dateiname */
              break;
          /*------------------------------------------------------------------*/
            case 'T':                   /*** "T": INSTALL (nie gesehen!) */
              t_desc="INSTALL=";        /* Bezeichnung DOS 4.0-MEM-kompatibel */
              break;
          /*------------------------------------------------------------------*/
            case 'F':                   /*** "F": Dateitabelle */
              t_desc="FILES=";          /* Bezeichnung DOS 4.0-MEM-kompatibel */
              break;
          /*------------------------------------------------------------------*/
            case 'X':                   /*** "X": FCB-Dateitabelle */
              t_desc="FCBS=";           /* Bezeichnung DOS 4.0-MEM-kompatibel */
              break;
          /*------------------------------------------------------------------*/
            case 'B':                   /*** "B": DOS-eigene Cache-Puffer */
            case 'C':                   /*** "C": Cache-Puffer im EMS */
              t_desc="BUFFERS=";        /* Bezeichnung DOS 4.0-MEM-kompatibel */
              break;
          /*------------------------------------------------------------------*/
            case 'L':                   /*** "L": Laufwerksumleitungen */
              t_desc="LASTDRIVE=";      /* Bezeichnung DOS 4.0-MEM-kompatibel */
              break;
          /*------------------------------------------------------------------*/
            case 'S':                   /*** "S": Stack-Bereiche */
              t_desc="STACKS=";         /* Bezeichnung DOS 4.0-MEM-kompatibel */
              break;
          /*------------------------------------------------------------------*/
            case 'e':                   /*** "e": EMS-Pageframe */
              t_name="EMS";             /* Bezeichnung */
              t_desc="EMS-Pageframe";
              break;
          /*------------------------------------------------------------------*/
            case 'r':                   /*** "r": BIOS-ROM */
            case 'x':
              t_name="Bios-ROM";        /* Teil des BIOS */
              t_desc=((*Bl)->type=='x')?"Erw. BIOS":"ROM-BIOS";
              break;
          /*------------------------------------------------------------------*/
            case 'v':                   /*** "v": Video-RAM */
              t_name="Video";
              t_desc="Videospeicher";
              break;
          /*------------------------------------------------------------------*/
            default:                    /*** unbekannter SCB */
              t_desc="??? (SCB ' ')";   /* Typ des Blocks ausgeben */
              t_desc[10]=(*Bl)->type;   /* Blockkennung eintragen */
          }
        }
        else {
          switch(psp=(*Bl)->psp) {        /* Blocktypen auswerten */
          /*------------------------------------------------------------------*/
            case 0x0000:                /*** Freies Segment */
              t_name="frei";            /* Dateiname: frei */
              t_desc="";                /* Keine Beschreibung */
              break;
          /*------------------------------------------------------------------*/
            case 0x0001:                /*** "verlorenes" Segment? */
              t_name="fehlt";           /* Besitzer unbekannt */
              t_desc="(unbekannt)";     /* Besitzer nicht mehr im Speicher */
              break;
          /*------------------------------------------------------------------*/
            case 0x0002:                /*** Segment nach MCBs */
              if((*Bl)->name[0]=='r')   /* Erluterung */
                t_name="ber MCB";
              else
                t_name="ohne RAM";      /* Erluterung */
              t_desc="(unbekannt)";     /* Besitzer nicht mehr im Speicher */
              break;
          /*------------------------------------------------------------------*/
            case 0x003:                 /*** Hauptspeicher <640 KB */
              t_desc="(unbekannt)";     /* Besitzer nicht mehr im Speicher */
              break;
          /*------------------------------------------------------------------*/
            case 0x004:                 /*** Fllsegment von NWDOS */
              t_name="Fll";
              t_desc="NWDOS Fllblock";
              break;
          /*------------------------------------------------------------------*/
            case 0x0006:                /*** DR-DOS: XMS-UMB */
              t_name="XMS umb";         /* Erluterung */
              sprintf(s,"gehrt %s",(*Bl)->name);
              t_desc=s;                 // Besitzer
              break;
          /*------------------------------------------------------------------*/
            case 0x0007:                /*** DR-DOS: RAM-Lcke */
              t_name="";        
              t_desc="DR-DOS excluded"; /* Erluterung */
              break;
          /*------------------------------------------------------------------*/
            case 0x0008:                /*** Vom System belegtes Segment */
              t_name="System";          /* Dateiname: System */
              if(t_adr==0x70)           /* System-Segment: Typ auswerten */
                t_desc="BetrSys BIOS";
              else if(t_adr==0xFFFF)    /* HMA */
                t_desc="DOS=HIGH";
              else if(_osmajor>=5 && strcmp((*Bl)->name,"SD")==0)
                t_desc="BetrSys Daten";
              else if(_osmajor>=5 && strcmp((*Bl)->name,"SC")==0)
                t_desc="BetrSys Programm";
              else
                t_desc="Betriebssystem";
              break;
          /*------------------------------------------------------------------*/
            default:
              if(psp==t_adr) {          /* Segment ist PSP-Segment? */
                t_name=(*Bl)->name;     /* Name steht im Block */
                t_desc="Programm";      /* Inhaltsbezeichnung */
              }
              else {                    /* Segment ist Datensegment */
                p2=GetBlkInfo(FirstBlkDesc,pspptr=_ptr(psp,0x0000),0);
                                        /* Besitzer-Block suchen */
                t_name=p2->name;        /* Name: Name des Besitzers */
                if(t_adr==pspptr->ENV)  /* Environment-Zeiger zeigt hierher? */
                  t_desc="Environment"; /* Dann kennzeichnen */
                else {
                  p=pspptr->PDOSIN;     /* Zeiger auf Dateitabelle */
                                        /* (von Funktion 21/67 ab DOS 3.3) */
                  if(FP_SEG(p)==t_adr && FP_OFF(p)==0)
                                        /* Zeiger zeigt hierher? */
                    t_desc="Handles";   /* Dann kennzeichnen */
                  else
                    t_desc="Daten";     /* Sonst irgendwas anderes */
                }
              }
          }
          t_chain=(char)(((*Bl)->sys==0)?' ':
                                         '');
                                        /* Zeichen MCB-Zugehrigkeit */
          if(opt_progcomp) {            /* Programme zusammenfassen? */
            count=1;                    /* Bislang erst ein Block */
            showadr=0;                  /* Noch kein PSP gefunden */
            while((*Bl)->next && (*Bl)->next->psp==psp) {
                                        /* Falls da: Folgende Blcke dazu */
              if((*Bl)->psp==(*Bl)->seg) {
                                        /* Segment ist PSP-Segment? */
                showadr=1;              /* Segmentadresse gltig */
                t_adr=psp;              /* PSP-Segment fr Anzeige merken */
              }
              (*Bl)=(*Bl)->next;        /* Block berspringen */
              MergeBlockLists(&(*sub),&((*Bl)->sub));
                                        /* Unterblcke zusammensortieren */
              MergeTechInfo(&info,&((*Bl)->info));
                                        /* Technische Infos zusammenwerfen */
              count++;                  /* Blcke zhlen */
              t_bytes+=(*Bl)->size;     /* Blockumfang hinzufgen */
            }
            if((*Bl)->psp==(*Bl)->seg) {/* Segment ist PSP-Segment? */
              showadr=1;                /* Segmentadresse gltig */
              t_adr=psp;                /* PSP-Segment fr Anzeige merken */
            }
            sprintf(t_desc=s,count==1?"in %d Block":"in %d Blcken",count);
                                        /* Blockzhler in Beschreibung */
            showadr|=(count==1);        /* Nur 1 Segment? Dessen Adr angeben */
            if(opt_progcomp && psp==0x0001)
              return 0;                 /* Zusammenfassung: "unbekannt" weg */
          }
          if((t_adr==0xFFFF && *t_name=='?') || level)
            t_name="belegt";            /* Unbekanter Name im HMA als "belegt */

        }

        if(level)                       /* In Unterebene Adresse einrcken */
          _cprintf(" ");
        if(showadr)                     /* Segmentadresse da? */
          if(t_adr!=0xFFFF)             /* Nicht HMA? */
            _cprintf("%04X",t_adr);     /* Dann diese anzeigen */
          else
            _cprintf("HMA ");           /* Sonst als HMA kennzeichnen */
        else
          _cprintf("    ");             /* Sonst nur ein bichen Platz */
        _cprintf("%c%-8s ",t_chain,t_name);
        if(level==0)                    /* Korrektur fr nicht-eingerckt */
          _cprintf(" ");
        if(t_bytes>-1)                  /* Gltige Lnge angegeben? */
          _cprintf("%6ld",t_bytes);     /* Dann Lnge ausgeben */
        else
          _cprintf("      ");           /* Sonst Lngenfeld leer lassen */
        if(level==0)                    /* Beschreibung normal/eingerckt */
          _cprintf(" %-16s ",t_desc);
        else
          _cprintf(" %c-%-14s ", ((*Bl)->next)?'':'',t_desc);

        if(opt_width>40)                /* >40 Zeichen/Zeile:Technische Infos */
          ListTechInfo(info,(*sub)!=NULL);

        _cprintf("\r\n");               /* Neue Zeile */
        return 1;                       /* Block wurde angezeigt */
}

/*----------------------------------------------------------------------------*/
void ListBlocks(struct MemBlDesc *p1,int level)
{
        int res;
        struct MemBlDesc *sub;

        for(;p1;p1=p1->next) {          /* Alle Blcke durchgehen */
          res=ListBlock(&p1,&sub,level);/* Block anzeigen */
          ListBlocks(sub,level+res);    /* ...diese (rekursiv) anzeigen */
        }                               
}

/*----------------------------------------------------------------------------*/
void SelectBlocks(struct MemBlDesc **List,struct TechInfo **Info)
{
        struct MemBlDesc *p1,*p2,
                         **prevnext;    /* Nachfolger-Verkettung im Vorgnger */
        char c,*name;
        unsigned psp;
        int count;
        int disp;

        prevnext=List;                  /* Adresse des Zeigers auf Listenkopf */
        p1=*List;                       /* Erstes Element */
        while(p1) {                     /* Ganze Liste durchgehen */
          SelectBlocks(&(p1->sub),&(p1->info));
                                        /* Erst mal Unterebene prfen */
          disp=1;                       /* default: Block anzeigen */
          if(p1->nopsp)  {
            c=p1->type;                 /* Blocktyp */
            if((!opt_showscb && strchr("DIFXBCLS",c)) ||
               (!opt_showhard && (c=='e' || c=='v')) ||
               (!opt_showbios && (c=='r' || c=='x')))
                                        /* Segment dieses Typs unterdrckt? */
              disp=0;                   /* Ja: weg damit */
          }
          else {
            psp=p1->psp;                /* Block-PSP */
            if((!opt_showfree && psp==0x0000) ||
               (!opt_showsys && psp==0x0008) ||
               (!opt_showother && (psp==0x0001 || psp==0x0002 || psp==0x0006)) ||
               (opt_progcomp && (psp==0x0007 || psp==0x0004)))
                                        /* Segment dieses Typs unterdrckt? */
              disp=0;                   /* Ja: weg damit */
            else
              if(!opt_showall && psp>0x0008) {
                name=GetBlkInfo(FirstBlkDesc,_ptr(psp,0x0000),0)->name;
                                        /* Namen des Besitzer-Blocks suchen */
                for(count=0;count<num_shownames;count++)
                                        /* Alle ausgewhlten Namen durch */
                if(!strcmp(shownames[count],name))
                  break;                /* Name in Liste? Dann Abbruch */
                if(count==num_shownames)/* Name nicht in Liste? */
                  disp=0;               /* Dann Block berspringen */
              }
          }
          if(!disp) {                   /* Block nicht zeigen? */
            if(Info)                    /* Hhere Ebene hat technische Infos? */
              MergeTechInfo(Info,&(p1->info));
                                        /* Technische Infos in hhere Ebene */
            *prevnext=p1->sub;          /* Verkettung Vorgnger auf Subkette */
            for(p2=p1->sub;p2;prevnext=&(p2->next),p2=p2->next)
              ;                         /* Ende der Subblock-Kette suchen */
            *prevnext=p1->next;         /* Verkettung von Subblock-Ende auf
                                           Nachfolger aktueller Block */
            free(p1);                   /* Weg mit dem Block */
            p1=*prevnext;               /* Weiter mit Block nach der Subkette */
          }
          else {
            prevnext=&(p1->next);       /* Nachfolger-Feld merken */
            p1=p1->next;                /* Zum nchsten Block */
          }
        }
}

/*----------------------------------------------------------------------------*/
void ListMemory(void)
{
        int count;

        if(opt_xmem)
          _cprintf("Konventioneller Speicher:\r\n");
        _cprintf(    "Seg  Name      Bytes  Beschreibung    ");
        if(opt_width>40)
          _cprintf(  " Technische Informationen");
        _cprintf("\r\n   ");
        if(opt_width>40) {
          _cprintf(" ");
          for(count=40;count<(int)opt_width;count++)
            _cprintf("");
        }
        _cprintf("\r\n");               /* Kopf der Tabelle ausgeben */
        SelectBlocks(&FirstBlkDesc,NULL);
                                        /* Unerwnschte Blcke aussortieren */
        ListBlocks(FirstBlkDesc,0);     /* Tabelle ausgeben */

        if(!opt_terse) {                /* Zahlen-Auswertung unterdrcken? */
          _cprintf("\r\n");             /* Neue Zeile */
          _cprintf("  Maximale Programmgre: %6ld Bytes  "
                   "  Programm-Ladesegment: %04X\r\n",
                   MaxSize,FP_SEG(MyPSP));/* Aktuelles PSP-Segment */
        }
}

/*----------------------------------------------------------------------------*/
void EMSInfo(void)
/*
   Ausgeben von Informationen ber den EMS-Speicher.
*/
{
        unsigned i;

        if(EMSvers!=-1) {               /* Nur, falls EMS berhaupt da */
          _cprintf("\r\nExpanded Memory/EMS (Version %d.%d):\r\n",
                   EMSvers>>4,EMSvers&0x0F);
          if(NumHandles) {              /* berhaupt Handle-Infos da? */
            _cprintf("Handle  Name     Bytes    Attribut\r\n"
                     "    \r\n");
            for(i=0;i<NumHandles;i++)   /* Handles durchgehen */
              _cprintf("%04X    %-8s %8ld %-12s\r\n",
                       HandleInfo[i].hdl,HandleInfo[i].name,
                       EMS_PAGE*HandleInfo[i].pages,
                       (HandleInfo[i].attr==0)?"volatile":
                       (HandleInfo[i].attr==1)?"non-volatile":
                                               "keines");
          }
          if(!opt_terse) {              /* Zahlen-Auswertung unterdrcken? */
            if(NumHandles) _cprintf("\r\n");
            _cprintf("  \"Normal\" belegt:%8ld Bytes        "
                     "  Als DOS-Speicher:%8ld Bytes\r\n",
                     TotalEMS-FreeEMS-DosEMS,DosEMS);
            _cprintf("  Freier Speicher:%8ld Bytes        "
                     "  Speicher total:%8ld Bytes\r\n",
                     FreeEMS,TotalEMS);
          }
        }
}

/*----------------------------------------------------------------------------*/
void XMSInfo(void)
/*
   Ausgeben von Informationen ber den XMS-Speicher.
*/
{
        unsigned i;

        if(XMSvers!=-1 || ExtTotal)     /* Nur, falls Extended Memory da */
          _cprintf("\r\nExtended Memory");
          if(XMSvers!=-1)               /* Nur, falls XMS da */
            _cprintf("/XMS (Version %x.%02x):\r\n",XMSvers>>8,XMSvers&0xFF);
          if(NumXMSHdls) {              /* berhaupt Handle-Infos da? */
            _cprintf("Handle Bytes    Lock count\r\n"
                     "  \r\n");
            for(i=0;i<NumXMSHdls;i++)   /* Handles durchgehen */
              _cprintf("%04X   %8ld    %3u\r\n",
                       XMSHdlInfo[i].hdl,
                       1024L*XMSHdlInfo[i].pages,
                       (unsigned)XMSHdlInfo[i].attr);
          }
          if(!opt_terse) {              /* Zahlen-Auswertung unterdrcken? */
            if(NumXMSHdls) _cprintf("\r\n");
            if(XMSvers!=-1) {           /* Nur, falls XMS da */
              _cprintf("  Extended ohne XMS:%8ld Bytes      "
                       "  XMS-Speicher frei:%8ld Bytes\r\n",
                       ExtTotal,FreeXMS);
            }
            else
              _cprintf(":\r\n"
                       "  Extended Memory:%8ld Bytes\r\n",ExtTotal);
          }
}
